import tornado.ioloop
import tornado.web
import liquer.server.handlers as h

class LiquerIndexHandler(h.LiquerIndexHandler, tornado.web.RequestHandler):
    pass

class LiquerJsHandler(h.LiquerIndexJsHandler, tornado.web.RequestHandler):
    pass
#/api/commands.json
class CommandsHandler(h.CommandsHandler, tornado.web.RequestHandler):
    pass

#/api/debug-json/<path:query>
class DebugQueryHandler(h.DebugQueryHandler, tornado.web.RequestHandler):
    pass

#/q/<path:query>
class QueryHandler(h.QueryHandler, tornado.web.RequestHandler):
    pass

class BuildHandler(h.BuildHandler, tornado.web.RequestHandler):
    pass

class RegisterCommandHandler(h.RegisterCommandHandler, tornado.web.RequestHandler):
    pass    

if __name__ == "__main__":
    import liquer.ext.basic
    import liquer.ext.meta
    import liquer.ext.lq_pandas
#    import liquer.ext.lq_hxl
    import liquer.ext.lq_python
    import liquer.ext.lq_pygments
    from liquer.state import set_var, get_vars
#    import liquer.commands
#    liquer.commands.enable_remote_registration()

    url_prefix='/liquer'
    port = 8080
    set_var("api_path",url_prefix+"/q/")
    set_var("server",f"http://localhost:{port}")
    

    application = tornado.web.Application([
        (r"/liquer/api/commands.json", CommandsHandler),
        (r"/liquer/api/debug-json/(.*)", DebugQueryHandler),
        (r"/liquer/api/build", BuildHandler),
        (r"/liquer/api/register_command/(.*)", RegisterCommandHandler),
        (r"/liquer/q/(.*)", QueryHandler),
        (r'/static/(.*)', tornado.web.StaticFileHandler, {'path': h.liquer_static_path()}),
        (r'/liquer/static/(.*)', tornado.web.StaticFileHandler, {'path': h.liquer_static_path()}),
        (r'/', LiquerIndexHandler),
        (r'/index.html', LiquerIndexHandler),
        (r'/liquer.js', LiquerJsHandler),
    ])
    application.listen(port)
    tornado.ioloop.IOLoop.current().start()



from enum import Enum

class Status(Enum):
    NONE = "none"
    EVALUATING_PARENT = "evaluating parent"
    EVALUATION = "evaluation"
    EVALUATING_DEPENDENCIES = "evaluating dependencies"
    FINISHED = "finished"
    READY = "ready"
    ERROR = "error"
    OBSOLETE = "obsolete"



import warnings
print ("WARNING: liquer.blueprint is deprecated, use liquer.server.blueprint")
warnings.warn("liquer.blueprint is deprecated, use liquer.server.blueprint", DeprecationWarning,
              stacklevel=2)
from liquer.server.blueprint import *



from os import makedirs, name, remove
from pathlib import Path
import json
from io import BytesIO

STORE=None
def get_store():
    global STORE
    if STORE is None:
        STORE = Store()
    return STORE

def set_store(store):
    global STORE
    STORE = store

class StoreException(Exception):
    def __init__(self, message, key=None, store=None):
        self.original_message=message
        if key is not None:
            message+=f":\n  key: {key}"
        if store is not None:
            message+=f"  store: {store}"

        super().__init__(message)
        self.key=key
        self.store=store

class KeyNotFoundStoreException(StoreException):
    def __init__(self, message="Key not found in store", key=None, store=None):
        super().__init__(message=message, key=key, store=store)

class KeyNotSupportedStoreException(StoreException):
    def __init__(self, message="Key not supported in store", key=None, store=None):
        super().__init__(message=message, key=key, store=store)

class StoreMixin:
    def with_overlay(self, overlay):
        return OverlayStore(overlay, self)
    def with_fallback(self, fallback):
        return OverlayStore(self, fallback)

class Store(StoreMixin):
    def parent_key(self, key):
        if key == "":
            return None
        return "/".join(key.split("/")[:-1])

    def key_name(self, key):
        return str(key.split("/")[-1])

    def default_metadata(self, key, is_dir=False):
        return dict(key=key, fileinfo=dict(name=self.key_name(key), is_dir=is_dir))

    def mount(self, key, store):
        return MountPointStore(self).mount(key, store)

    def get_bytes(self, key):
        raise KeyNotFoundStoreException(key=key, store=self)

    def get_metadata(self, key):
        raise KeyNotFoundStoreException(key=key, store=self)

    def store(self, key, data, metadata):
        raise KeyNotSupportedStoreException(key=key, store=self)

    def store_metadata(self, key, metadata):
        raise KeyNotSupportedStoreException(key=key, store=self)

    def remove(self, key):
        raise KeyNotFoundStoreException(key=key, store=self)

    def removedir(self, key):
        raise KeyNotFoundStoreException(key=key, store=self)

    def contains(self, key):
        return key == ""

    def is_dir(self, key):
        return key == ""

    def keys(self):
        return []

    def listdir(self, key):
        return []

    def makedir(self, key):
        raise KeyNotSupportedStoreException(key=key, store=self)

    def openbin(self, key, mode="r", buffering=-1):
        raise KeyNotSupportedStoreException(key=key, store=self)
    
    def is_supported(self, key):
        return False

    def __str__(self):
        return f"Empty store"

    def __repr__(self):
        return f"Store()"

class FileStore(Store):
    METADATA = "__metadata__"
    def __init__(self, path):
        if isinstance(path,Path):
            self.path = path
        else:
            self.path = Path(path)
    
    def path_for_key(self, key):
        if key in (None,""):
            return self.path
        p = self.path / key
        assert p.name != self.METADATA            
        return p

    def metadata_path_for_key(self, key):
        p = self.path / key
        assert p.name != self.METADATA
        return p.parent / self.METADATA / (p.name + ".json")

    def get_bytes(self, key):
        if not self.path_for_key(key).exists():
            raise KeyNotFoundStoreException(key=key, store=self)
        return open(self.path_for_key(key), "rb").read()

    def get_metadata(self, key):
        p = self.path_for_key(key)
        metadata = self.default_metadata(key, p.is_dir())
        if p.is_dir():
            return metadata
        else:
            if self.path_for_key(key).exists():
                if self.metadata_path_for_key(key).exists():
                    metadata.update(json.loads(open(self.metadata_path_for_key(key)).read()))
            else:
                if self.metadata_path_for_key(key).exists():
                    metadata.update(json.loads(open(self.metadata_path_for_key(key)).read()))
                else:
                    raise KeyNotFoundStoreException(key=key, store=self)
            metadata["key"] = key
            metadata["fileinfo"]=dict(name=str(p.name), is_dir = False)
        return metadata

    def store(self, key, data, metadata):
        self.path_for_key(key).parent.mkdir(parents=True, exist_ok=True)
        self.path_for_key(key).write_bytes(data)
        self.store_metadata(key, metadata)

    def store_metadata(self, key, metadata):
        self.metadata_path_for_key(key).parent.mkdir(parents=True, exist_ok=True)
        with open(self.metadata_path_for_key(key), "w") as f:
            json.dump(metadata, f)

    def remove(self, key):
        self.path_for_key(key).unlink(missing_ok=True)
        self.metadata_path_for_key(key).unlink(missing_ok=True)

    def removedir(self, key):
        (self.path_for_key(key) / self.METADATA).rmdir()
        self.path_for_key(key).rmdir()

    def contains(self, key):
        return self.path_for_key(key).exists()

    def is_dir(self, key):
        return self.path_for_key(key).is_dir()

    def keys(self, parent=None):
        d = self.listdir(parent)
        if d is None:
            return []
        else:
            for k in d:
                key = k if parent is None else parent + "/" + k
                yield key
                for kk in self.keys(key):
                    yield kk

    def listdir(self, key):
        if self.is_dir(key):
            return [d.name for d in self.path_for_key(key).iterdir() if d.name != self.METADATA]
            
    def makedir(self, key):
        self.path_for_key(key).mkdir(parents=True, exist_ok=True)
        (self.path_for_key(key) / self.METADATA).mkdir(parents=True, exist_ok=True)

    def openbin(self, key, mode="rb", buffering=-1):
        mode = dict(r="rb", w="wb").get(mode, mode)
        return open(self.path_for_key(key), mode)

    def is_supported(self, key):
        return True

    def __str__(self):
        return f"File store at {self.path}"

    def __repr__(self):
        return f"FileStore('{self.path}')"


class MemoryStore(Store):
    def __init__(self):
        self.directories = set()
        self.data = {}
        self.metadata = {}
    
    def get_bytes(self, key):
        if key not in self.data:
            raise KeyNotFoundStoreException(key=key, store=self)
        return self.data[key]

    def get_metadata(self, key):
        if self.is_dir(key):
            if key in ("",None):
                return dict(key="", fileinfo=dict(name="", is_dir=True))
            else:
                return dict(key=str(key), fileinfo=dict(name=str("/".split(key)[-1]), is_dir=True))
        else:
            if key not in self.metadata:
                raise KeyNotFoundStoreException(key=key, store=self)
            return self.metadata[key]

    def store(self, key, data, metadata):
        self.makedir(self.parent_key(key))
        self.data[key]=data
        self.metadata[key]=metadata

    def store_metadata(self, key, metadata):
        self.metadata[key]=metadata

    def remove(self, key):
        try:
            self.directories.remove(key)
        except KeyError:
            pass
        try:
            del self.data[key]
        except KeyError:
            pass
        try:
            del self.metadata[key]
        except KeyError:
            pass

    def removedir(self, key):
        if len(self.listdir(key))==0:
            try:
                self.directories.remove(key)
            except KeyError:
                pass

    def contains(self, key):
        return key in self.directories or key in self.data or key in self.metadata

    def is_dir(self, key):
        if key in (None, ""):
            return True
        return key in self.directories

    def keys(self):
        k = set()
        k.update(self.directories)
        k.update(self.data.keys())
        k.update(self.metadata.keys())
        return sorted(k)

    def listdir(self, key):
        if key in ("",None):
            return sorted(set(k.split("/")[0] for k in self.keys()))
        if self.is_dir(key):
            return [self.key_name(k) for k in self.keys() if self.parent_key(k)==key]
            
    def makedir(self, key):
        while key not in (None, ""):
            self.directories.add(key)
            key = self.parent_key(key)

    def openbin(self, key, mode="r", buffering=-1):
        mode = dict(r="rb", w="wb").get(mode, mode)
        if mode == "rb":
            return BytesIO(self.data[key])
        raise Exception("openbin not supported for write yet")

    def is_supported(self, key):
        return True

    def __str__(self):
        return f"Memory store"

    def __repr__(self):
        return f"MemoryStore()"

class OverlayStore(Store):
    """Overlay store combines two stores: overlay and fallback.
    Overlay is used as a primary store for reading and writing.
    The fallback is used only for reading if key is not found in the overlay store (and is not removed).
    Thus the fallback store is never modified.
    """
    def __init__(self, overlay, fallback):
        self.overlay = overlay
        self.fallback = fallback
        self.removed=set()
    
    def get_bytes(self, key):
        if key not in self.removed:
            if self.overlay.contains(key):
                return self.overlay.get_bytes(key)
            else:
                return self.fallback.get_bytes(key)

    def get_metadata(self, key):
        if key not in self.removed:
            if self.overlay.contains(key):
                return self.overlay.get_metadata(key)
            else:
                return self.fallback.get_metadata(key)

    def store(self, key, data, metadata):
        try:
            self.removed.remove(key)
        except KeyError:
            pass
        return self.overlay.store(key, data, metadata)

    def store_metadata(self, key, metadata):
        try:
            self.removed.remove(key)
        except KeyError:
            pass
        return self.overlay.store_metadata(key, metadata)

    def remove(self, key):
        if key not in self.removed:
            if self.overlay.contains(key):
                self.overlay.remove(key)
            if self.fallback.contains(key):
                self.removed.add(key)

    def removedir(self, key):
        if self.contains(key):
            if len(self.listdir(key))==0:
                if self.overlay.contains(key):
                    self.overlay.remove(key)
                else:
                    self.removed.add(key)

    def contains(self, key):
        if key in self.removed:
            return False
        else:
            return self.overlay.contains(key) or self.fallback.contains(key)

    def is_dir(self, key):
        if key in self.removed:
            return False
        else:
            if self.overlay.contains(key):
                return self.overlay.is_dir(key)
            else:
                return self.fallback.is_dir(key)

    def keys(self):
        return sorted(set(self.overlay.keys()).union(self.fallback.keys()).difference(self.removed))

    def listdir(self, key):
        if key.endswith("/"):
            key = key[:-1]
        ld = self.overlay.listdir(key)
        d = set() if ld is None else set(ld)
        ld = self.fallback.listdir(key)
        d = d if ld is None else d.union(ld)
        return [x for x in sorted(d) if key+"/"+x not in self.removed]
            
    def makedir(self, key):
        if key in self.removed:
            self.removed.remove(key)
        return self.overlay.makedir(key)

    def openbin(self, key, mode="r", buffering=-1):
        mode = dict(r="rb", w="wb").get(mode, mode)
        if mode == "rb":
            if self.overlay.contains(key):
                return self.overlay.openbin(key, mode, buffering)
            elif self.fallback.contains(key):
                return self.fallback.openbin(key, mode, buffering)
        elif mode == "wb":
            return self.overlay.openbin(key, mode, buffering)

    def is_supported(self, key):
        return self.overlay.is_supported(key) or self.fallback.is_supported(key())

    def __str__(self):
        return f"Overlay of ({self.overlay}) over ({self.fallback})"

    def __repr__(self):
        return f"OverlayStore({repr(self.overlay)},{repr(self.fallback)})"

class RoutingStore(Store):
    def route_to(self, key):
        raise KeyNotSupportedStoreException(key=key, store=self)

    def is_supported(self, key):
        try:
            store = self.route_to(key)
            if store is None:
                return False
            return store.is_supported(key)
        except KeyNotSupportedStoreException:
            return False

    def get_bytes(self, key):
        return self.route_to(key).get_bytes(key)

    def get_metadata(self, key):
        return self.route_to(key).get_metadata(key)

    def store(self, key, data, metadata):
        return self.route_to(key).store(key, data, metadata)

    def store_metadata(self, key, metadata):
        return self.route_to(key).store_metadata(key, metadata)

    def remove(self, key):
        return self.route_to(key).remove(key)

    def removedir(self, key):
        return self.route_to(key).removedir(key)

    def contains(self, key):
        return self.route_to(key).contains(key)

    def is_dir(self, key):
        return self.route_to(key).is_dir(key)

    def keys(self):
        raise NotImplementedError

    def listdir(self, key):
        return self.route_to(key).listdir(key)
            
    def makedir(self, key):
        return self.route_to(key).makedir(key)

    def openbin(self, key, mode="r", buffering=-1):
        return self.route_to(key).openbin(key, mode=mode, buffering=buffering)

    def __str__(self):
        return f"Abstract routing store"

    def __repr__(self):
        return f"RoutingStore()"


class KeyTranslatingStore(Store):
    def __init__(self,store):
        self.substore = store

    def translate_key(self, key, inverse=False):
        return key

    def is_supported(self, key):
        try:
            tkey = self.translate_key(key)
            if tkey is None:
                return False
            return self.substore.is_supported(tkey)
        except KeyNotSupportedStoreException:
            return False

    def get_bytes(self, key):
        return self.substore.get_bytes(self.translate_key(key))

    def get_metadata(self, key):
        return self.substore.get_metadata(self.translate_key(key))

    def store(self, key, data, metadata):
        return self.substore.store(self.translate_key(key), data, metadata)

    def store_metadata(self, key, metadata):
        return self.substore.store_metadata(self.translate_key(key), metadata)

    def remove(self, key):
        return self.substore.remove(self.translate_key(key))

    def removedir(self, key):
        return self.substore.removedir(self.translate_key(key))

    def contains(self, key):
        return self.substore.contains(self.translate_key(key))

    def is_dir(self, key):
        return self.substore.is_dir(self.translate_key(key))

    def keys(self):
        for key in self.substore.keys():
            yield self.translate_key(key, inverse=True)

    def listdir(self, key):
        return self.substore.listdir(self.translate_key(key))
            
    def makedir(self, key):
        return self.substore.makedir(self.translate_key(key))

    def openbin(self, key, mode="r", buffering=-1):
        return self.substore.openbin(self.translate_key(key), mode=mode, buffering=buffering)

    def __str__(self):
        return f"Key translating store on ({self.substore})"

    def __repr__(self):
        return f"KeyTranslatingStore({repr(self.substore)})"

class PrefixStore(KeyTranslatingStore):
    def __init__(self,store, prefix):
        self.substore = store
        self.prefix = prefix

    def translate_key(self, key, inverse=False):
        prefix = self.prefix+"/"
        if inverse:
            if key in (None, ""):
                return self.prefix
            else:
                return prefix + key
        else:
            if key  == self.prefix:
                return ""
            else:
                if key.startswith(prefix):
                    return key[len(prefix):]
                else:
                    raise KeyNotSupportedStoreException

    def contains(self, key):
        if key == self.prefix:
            return True
        return self.substore.contains(self.translate_key(key))

    def is_dir(self, key):
        if key == self.prefix:
            return True
        return self.substore.is_dir(self.translate_key(key))

    def __str__(self):
        return f"Prefixed store {self.prefix} on ({self.substore})"

    def __repr__(self):
        return f"PrefixStore({repr(self.substore)}, prefix={repr(self.prefix)})"

class MountPointStore(RoutingStore):
    def __init__(self, default_store=None, routing_table=None):
        self.default_store = default_store
        self.routing_table=[] if routing_table is None else routing_table

    def mount(self, key, store):
        self.routing_table.append((key, PrefixStore(store, prefix=key)))
        return self

    def route_to(self, key):
        for prefix, store in reversed(self.routing_table):
            if key == prefix:
                return store
            if not prefix.endswith("/"):
                prefix+="/"
            if key.startswith(prefix):
                if store.is_supported(key):
                    return store
        if self.default_store is not None:
            return self.default_store

    def keys(self):
        prefixes=[]
        for prefix, store in reversed(self.routing_table):
            yield prefix
            if not prefix.endswith("/"):
                prefix+="/"
            for key in store.keys():
                if any(key.startswith(p) for p in prefixes):
                    continue
                if key.startswith(prefix):
                    yield key 
            prefixes.append(prefix)
        if self.default_store is not None:
            for key in self.default_store.keys():
                if any(key.startswith(p) for p in prefixes):
                    continue
                yield key

    def listdir(self, key):
        d = set(self.route_to(key).listdir(key))
        key_split = key.split("/")
        if len(key_split)== 1 and key_split[0]=="":
            key_split=[]
        key_depth = len(key_split)
        for prefix, _ in self.routing_table:
            if prefix.startswith(key+"/") or key in (None, ""):
                v = prefix.split("/")
                d.add(v[key_depth])
        return sorted(d)

    def removedir(self, key):
        for k,store in self.routing_table:
            if k == key:
                raise StoreException(f"Can't remove {key} because it is a mount point of {repr(store)}", key=key, store=self)

        return self.route_to(key).removedir(key)

    def __str__(self):
        return f"Mount point store routed by {self.routing_table} with default store {self.default_store}"

    def __repr__(self):
        return f"MountPointStore({repr(self.default_store)}, routing_table={repr(self.routing_table)})"

class FileSystemStore(Store):
    METADATA = "__metadata__"
    def __init__(self, fs, path=""):
        if isinstance(path,Path):
            self.path = path
        else:
            self.path = Path(path)
        self.fs = fs
    
    def path_for_key(self, key):
        if key in (None,""):
            return str(self.path)
        p = self.path / key
        assert p.name != self.METADATA            
        return str(p)

    def metadata_path_for_key(self, key):
        p = self.path / key
        assert p.name != self.METADATA
        return str(p.parent / self.METADATA / (p.name + ".json"))

    def metadata_dir_path_for_key(self, key):
        p = self.path / key
        assert p.name != self.METADATA
        return str(p.parent / self.METADATA)

    def get_bytes(self, key):
        if not self.fs.exists(self.path_for_key(key)):
            raise KeyNotFoundStoreException(f"Can't find {self.path_for_key(key)} in filesystem {self.fs}", key=key, store=self)
        return self.fs.readbytes(self.path_for_key(key))

    def get_metadata(self, key):
        p = self.path_for_key(key)
        isdir = self.fs.isdir(p)
        metadata = self.default_metadata(key, isdir)
        if isdir:
            return metadata
        else:
            if self.fs.exists(self.path_for_key(key)):
                if self.fs.exists(self.metadata_path_for_key(key)):
                    metadata.update(json.loads(self.fs.readtext(self.metadata_path_for_key(key))))
            else:
                if self.fs.exists(self.metadata_path_for_key(key)):
                    metadata.update(json.loads(self.fs.readtext(self.metadata_path_for_key(key))))
                else:
                    raise KeyNotFoundStoreException(key=key, store=self)
            metadata["key"] = key
            metadata["fileinfo"]=dict(name=str(self.key_name(key)), is_dir = False)
        return metadata

    def store(self, key, data, metadata):
        self.fs.makedirs(self.path_for_key(self.parent_key(key)))
        self.fs.writebytes(self.path_for_key(key), data)
        assert self.fs.exists(self.path_for_key(key))
        self.store_metadata(key, metadata)

    def store_metadata(self, key, metadata):
        parent = self.metadata_dir_path_for_key(key)
        self.fs.makedirs(parent)
        with self.fs.open(self.metadata_path_for_key(key), "w") as f:
            json.dump(metadata, f)

    def remove(self, key):
        if not self.contains(key):
            raise KeyNotFoundStoreException(key=key, store=self)
        if self.is_dir(key):
            raise KeyNotFoundStoreException(message=f"Can't remove a directory {key}; use removedir", key=key, store=self)
        try:
            self.fs.remove(self.path_for_key(key))
        except:
            pass
        try:
            self.fs.remove(self.metadata_path_for_key(key))
        except:
            pass

    def removedir(self, key):
        metadir = self.path_for_key(key) + "/" + self.METADATA
        try:
            self.fs.removedir(metadir)
        except:
            pass
        self.fs.removedir(self.path_for_key(key))            

    def contains(self, key):
        return self.fs.exists(self.path_for_key(key))

    def is_dir(self, key):
        return self.fs.isdir(self.path_for_key(key))

    def keys(self, parent=None):
        d = self.listdir(parent)
        if d is None:
            return []
        else:
            for k in d:
                key = k if parent is None else parent + "/" + k
                yield key
                for kk in self.keys(key):
                    yield kk

    def listdir(self, key):
        if self.is_dir(key):
            return [self.key_name(d) for d in self.fs.listdir(self.path_for_key(key)) if self.key_name(d) != self.METADATA]
            
    def makedir(self, key):
        self.fs.makedir(self.path_for_key(key), recreate=True)
        self.fs.makedir(self.path_for_key(key) + "/" +self.METADATA, recreate=True)

    def openbin(self, key, mode="rb", buffering=-1):
        return self.fs.openbin(self.path_for_key(key), mode=mode, buffering=buffering)

    def is_supported(self, key):
        return True

    def __str__(self):
        return f"Filesystem {self.fs} store at {self.path}"

    def __repr__(self):
        return f"FileSystemStore({repr(self.fs)}, {repr(self.path)})"



from io import StringIO, BytesIO
from tempfile import mkstemp
import os

from keras.models import Model, model_from_json, model_from_yaml, load_model, clone_model
from keras.utils import plot_model, print_summary

from liquer.state_types import StateType, register_state_type, mimetype_from_extension
from liquer.commands import command, first_command
from liquer.parser import encode, decode
from liquer.query import evaluate
from liquer.state import State

class KerasModelStateType(StateType):
    def identifier(self):
        return "kerasmodel"

    def default_extension(self):
        return "h5"

    def is_type_of(self, data):
        return isinstance(data, Model)

    def as_bytes(self, data, extension=None):
        if extension is None:
            extension = self.default_extension()
        assert self.is_type_of(data)
        mimetype = mimetype_from_extension(extension)
        if extension == "json":
            output = StringIO()
            output.write(data.to_json())
            return output.getvalue().encode("utf-8"), mimetype
        elif extension == "yaml":
            output = StringIO()
            output.write(data.to_yaml())
            return output.getvalue().encode("utf-8"), mimetype
        elif extension in ("h5", "hdf5"):
            handle, name = mkstemp(prefix="keras_model_",suffix="."+extension) # HACK - we need a file name, NamedTemporaryFile implementation does not work in windows
            os.close(handle)
            data.save(name)            
            b = open(name,"rb").read()
            os.remove(name)
            return b, mimetype
        else:
            raise Exception(
                f"Serialization: file extension {extension} is not supported by kerasmodel type.")

    def from_bytes(self, b: bytes, extension=None):
        if extension is None:
            extension = self.default_extension()

        if extension == "json":
            return model_from_json(b)
        elif extension == "yaml":
            return model_from_yaml(b)
        elif extension in ["h5", "hdf5"]:
            handle, name = mkstemp(prefix="keras_model_",suffix="."+extension) # HACK - we need a file name, NamedTemporaryFile implementation does not work in windows
            os.close(handle)
            with open(name, "wb") as f:
                f.write(b)
            model = load_model(f.name)
            os.remove(name)
            return model
        raise Exception(
            f"Deserialization: file extension {extension} is not supported by kerasmodel type.")

    def copy(self, data):
        model = clone_model(data)
        model.set_weights(data.get_weights())
        return model


KERASMODEL_STATE_TYPE = KerasModelStateType()
register_state_type(Model, KERASMODEL_STATE_TYPE)

@command
def keras_plot_model(model, show_shapes:bool=False, show_layer_names:bool=True, rankdir:str='TB', expand_nested:bool=False, dpi:int=96):
    "Keras plot model as png"
    assert isinstance(model, Model)
    handle, name = mkstemp(prefix="keras_model_",suffix=".png") # HACK - we need a file name, NamedTemporaryFile implementation does not work in windows
    os.close(handle)
    plot_model(model, to_file=name, show_shapes=show_shapes, show_layer_names=show_layer_names, rankdir=rankdir, expand_nested=expand_nested, dpi=96)
    b = open(name,"rb").read()
    return b

@command
def keras_summary(model):
    "Keras model summary"
    s = StringIO()
    print_summary(model, print_fn=lambda x,s=s:s.write(str(x)+"\n") )
    return s.getvalue()


import json
from liquer.state_types import (
    type_identifier_of,
    copy_state_data,
    mimetype_from_extension,
)
from copy import deepcopy
from liquer.parser import QueryException, Position

_vars = None

class EvaluationException(QueryException):
    pass


def get_vars():
    global _vars
    if _vars is None:
        _vars = {}
    return _vars


def vars_clone():
    """Return a copy of state variables dictionary
    This is used to set the initial content of State.vars for a newly created state.
    """
    return deepcopy(get_vars())


def set_var(name, value):
    """Set initial value of a state variable
    This is used to configure initial content of the state variables.

    It should be done before any query is evaluated.
    Calls at a later stage may lead to unintended behaviour like
    cache inconsistencies (query result may depend on initial value of a state variable,
    but this is not reflected in a query string).
    """
    get_vars()
    _vars[name] = value


class State(object):
    def __init__(self, data=None, metadata=None, context=None):
        self.data = data
        self.exception=None
        self.context=context
        self.metadata = metadata if metadata is not None else dict(
            query="",
            sources=[],
            log=[],
            is_error=False,
            vars=vars_clone(),
            filename=None,
            extension=None,
            message="",
            commands=[],
            extended_commands=[],
            type_identifier=None,
            caching=True,
            attributes={},
        )

    def next_state(self):
        state = self.__class__()
        state = state.from_dict(self.as_dict()).with_data(None)
        return state

    @property
    def query(self):
        return self.metadata["query"]

    @query.setter
    def query(self, value):
        self.metadata["query"] = value

    @property
    def is_error(self):
        return self.metadata["is_error"]

    @is_error.setter
    def is_error(self, value):
        self.metadata["is_error"] = value

    @property
    def vars(self):
        return self.metadata["vars"]

    @vars.setter
    def vars(self, value):
        self.metadata["vars"] = value

    @property
    def type_identifier(self):
        return self.metadata["type_identifier"]

    @type_identifier.setter
    def type_identifier(self, value):
        self.metadata["type_identifier"] = value


#    def with_caching(self, caching=True):
#        """Enables or disables caching for this state"""
#        # TODO: Make sure caching is propagated to dependent states
#        # TODO: Examples and documentation
#        print(f"QUERY {self.query}\nWITH CACHING {caching}")
#        if self.context is not None:
#            self.context.caching = caching
#            print(f"set context {id(self.context)} caching {self.context.caching}")
#        self.metadata["caching"] = caching
#        return self

    def with_data(self, data):
        """Set the data"""
        self.data = data
        self.metadata["type_identifier"] = type_identifier_of(data)
        return self

    def with_source(self, source):
        """Add sources to the state"""
        self.metadata["sources"] = [source] + self.metadata["sources"]
        return self

    def is_volatile(self):
        return self.metadata.get("attributes",{}).get("volatile", False)

    def set_volatile(self, flag):
        self.metadata["attributes"]["volatile"] = flag
        return self
    
    def get(self):
        """Get data from the state"""
        if self.is_error:
            tb = "\n".join(m.get("traceback", "") for m in self.metadata["log"])
            print(tb)
            position=None
            query=None
            for entry in self.metadata["log"]:
               if entry.get("kind")=="error":
                   position = Position.from_dict(entry.get("position"))
                   query=entry.get("query")

            if self.exception is None:
                raise EvaluationException(tb+"\n"+self.metadata["message"],position=position, query=query)
            else:
                raise self.exception

        return self.data

    def log_command(self, qv, number):
        """Log a command"""
        if self.context is None:
            self.metadata["log"].append(dict(kind="command", qv=qv, command_number=number))
        else:
            self.context.log_action(qv,number)
        return self

    def log_error(self, message):
        """Log an error message"""
        if self.context is None:
            self.metadata["log"].append(dict(kind="error", message=message))
            self.is_error = True
            self.metadata["message"] = message
        else:
            self.context.error(message)
        return self

    def log_warning(self, message):
        """Log a warning message"""
        if self.context is None:
            self.metadata["log"].append(dict(kind="warning", message=message))
            self.metadata["message"] = message
        else:
            self.context.warning(message)
        return self

    def log_exception(self, message, traceback):
        """Log an exception"""
        if self.context is None:
            self.metadata["log"].append(
                dict(kind="error", message=message, traceback=traceback)
            )
            self.is_error = True
            self.metadata["message"] = message
        else:
            self.context.exception(message, traceback)
        return self

    def log_info(self, message):
        """Log a message (info)"""
        if self.context is None:
            self.metadata["log"].append(dict(kind="info", message=message))
            self.metadata["message"] = message
        else:
            self.context.info(message)
        return self

    def as_dict(self):
        """Represent state metadata as a dictionary - suitable for json serialization
        State data is NOT part of the returned dictionary.
        """
        return deepcopy(self.metadata)

    def mimetype(self):
        """Return mime type of the data"""
        return mimetype_from_extension(self.extension)

    def from_dict(self, metadata):
        """Fill state by valueas from a dictionary"""
        if isinstance(metadata, self.__class__):
            metadata = metadata.metadata
        self.metadata = deepcopy(metadata)
        return self

    def has_flag(self, name):
        return self.metadata["vars"].get(name) == True

    def clone(self):
        """Clone the state including the deep copy of the data"""
        state = self.__class__()
        state = state.from_dict(self.as_dict())
        state.data = copy_state_data(self.data)
        return state

    def with_filename(self, filename):
        """set filename"""
        self.metadata["filename"] = filename
        if "." in filename:
            self.metadata["extension"] = filename.split(".")[-1].lower()
        return self



from notebook.utils import url_path_join
from notebook.base.handlers import IPythonHandler
import liquer.server.handlers as h
import os.path

class LiquerIndexHandler(IPythonHandler):
    def prepare(self):
        header = "Content-Type"
        body = "text/html"
        self.set_header(header, body)

    def get(self):
        self.write(open(os.path.join(h.liquer_static_path(),"index.html")).read())

class LiquerJsHandler(IPythonHandler):
    def prepare(self):
        header = "Content-Type"
        body = "text/javascript"
        self.set_header(header, body)

    def get(self):
        self.write(open(os.path.join(h.liquer_static_path(),"liquer.js")).read())

#/api/commands.json
class CommandsHandler(h.CommandsHandler, IPythonHandler):
    pass

#/api/debug-json/<path:query>
class DebugQueryHandler(h.DebugQueryHandler, IPythonHandler):
    pass

#/q/<path:query>
class QueryHandler(h.QueryHandler, IPythonHandler):
    pass

class BuildHandler(h.BuildHandler, IPythonHandler):
    pass

class RegisterCommandHandler(h.RegisterCommandHandler, IPythonHandler):
    pass    

def load_jupyter_server_extension(nb_server_app):
    """
    Called when the extension is loaded.

    Args:
        nb_server_app (NotebookWebApplication): handle to the Notebook webserver instance.
    """

    import liquer.ext.basic
    import liquer.ext.meta
    import liquer.ext.lq_pandas
    #import liquer.ext.lq_hxl
    import liquer.ext.lq_python
    import liquer.ext.lq_pygments
    from liquer.state import set_var, get_vars
    import liquer.commands

    liquer.commands.enable_remote_registration()

    web_app = nb_server_app.web_app
    host_pattern = '.*$'
    url_prefix='/liquer'
    route_pattern = url_path_join(web_app.settings['base_url'], url_prefix)

    set_var("api_path","/q/")
    set_var("server",route_pattern)

    web_app.add_handlers(host_pattern, [
        (route_pattern, LiquerIndexHandler),
        (url_path_join(route_pattern,'/index.html'), LiquerIndexHandler),
        (url_path_join(route_pattern,'/liquer.js'), LiquerJsHandler),
        (url_path_join(route_pattern,'/static/index.html'), LiquerIndexHandler),
        (url_path_join(route_pattern,'/static/liquer.js'), LiquerJsHandler),
        (url_path_join(route_pattern,'/api/commands.json'), CommandsHandler),
        (url_path_join(route_pattern,'/api/debug-json/(.*)'), DebugQueryHandler),
        (url_path_join(route_pattern,'/api/build'), BuildHandler),
        (url_path_join(route_pattern,'/api/register_command/(.*)'), RegisterCommandHandler),
        (url_path_join(route_pattern,'/q/(.*)'), QueryHandler),
    ])



from liquer.commands import command, first_command, command_registry
from liquer.query import evaluate_template
import liquer.ext.basic

@first_command(ns="meta")
def commands():
    """Returns a list of commands in json format"""
    return command_registry().as_dict()

@first_command(ns="meta")
def flat_commands(*pairs):
    """Returns a list of commands in json format"""
    filters = [(pairs[i-1],pairs[i]) for i in range(1,len(pairs),2)]
    d=[]
    for rl in command_registry().as_dict().values():
        for r in rl.values():
            if len(filters)==0 or all(r["attributes"].get(key) == value for key,value in filters):
                c={key:r.get(key) for key in ["name", "label", "module", "doc"]}
                for key, value in r["attributes"].items():
                    if key not in c:
                        c[key]=value
                d.append(c)
    return d

@first_command(ns="meta")
def flat_commands_nodoc(*pairs):
    """Returns a list of commands in json format"""
    filters = [(pairs[i-1],pairs[i]) for i in range(1,len(pairs),2)]
    d=[]
    for rl in command_registry().as_dict().values():
        for r in rl.values():
            if len(filters)==0 or all(r["attributes"].get(key) == value for key,value in filters):
                c={key:r.get(key) for key in ["name", "label", "module"]}
                for key, value in r["attributes"].items():
                    if key not in c and key!="context_menu":
                        c[key]=value
                d.append(c)
    return d

@command
def state(state):
    """Returns dictionary with all metadata for the resulting state"""
    return state.as_dict()

@command(ns="meta")
def help(state, command_name, ns="root"):
    """Returns a description of the command"""
    crd = command_registry().as_dict()
    make_link =  ("py" in crd) and ("pygments" in crd)
    try:
        r = crd[ns]
        try:
            c = r[command_name]
            html=f"<html><head><title>{command_name}</title></head>"
            html="<body\n>"
            html=f"<h1>{command_name} <em>({ns})</em></h1>\n"
            html+=f"<pre>{c['doc']}</pre>\n"
            html+="<h3>Arguments</h3>\n"
            html+="  <ul>\n"
            for a in c["arguments"]:
                html+=f"    <li><b>{a['name']}</b> - type: {a['type']}</li>\n"
            html+="  </ul>"
            html+="<h3>Attributes</h3>\n"
            html+="  <ul>\n"
            for key,value in sorted(c["attributes"].items()):
                if key == "example_link":
                    q=value
                    url=state.vars.get("server", "http://localhost")+state.vars.get("api_path", "/q/")+q
                    html+=f'    <li><b>{key}</b>:<a href="{url}">{value}</a></li>\n'
                else:
                    html+=f"    <li><b>{key}</b>:{value}</li>\n"
            html+="  </ul>"
            html+=f"<h3>Python module</h3>\n"
            module = c['module']
            if make_link:
                modfile = module.replace(".","_")
                html+=evaluate_template(f'<a href="$ns-py-pygments/module_source-{module}/highlight/link-url-html$/{modfile}.html">{module}</a>\n')
            else:
                html+=module

            html+="</body>"
            html+="</html>"
            return state.with_data(html).with_filename("help.html")
        except KeyError:
            html = f"<h1>Command <em>{command_name}</em> not registered in namespace <em>{ns}</em></h1>"
            candidates = ", ".join(key for key,value in crd.items() if command_name in value)
            if len(candidates):
                html+=f"Command can be found in: {candidates}"
            else:
                html+="There is no such command in available namespaces"
            return state.with_data(html).with_filename("help.html")
    except KeyError:
        return state.with_data(f"<h1>Unknown namespace <em>{ns}</em></h1>").with_filename("help.html")

if __name__=="__main__":
    from liquer import *
    print(evaluate("ns-meta/flat_commands").get())
    #print(evaluate("ns-meta/commands/state").get()["commands"])
    #evaluate_and_save("ns-meta/commands/state/state.json")